//////////////////////////////////////////////////////////////////////////
// File     : 2226demo.c						//
// Function : Linux demo application for Sensoray model 2226 board	//
// Author   : Jim Lamberson, Charlie Liu, Pete Eberlein, D.A.		//
// Revision :								//
//            Aug., 2009   Pete Eberlein  Revised for 2226		//
//									//
//////////////////////////////////////////////////////////////////////////

/*
Copyright (C) 2009-2010  Sensoray Co., Inc.  This program is free software;
you can redistribute it and/or modify it under the terms of the GNU
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

You must obey the GNU General Public License in all respects for all of
the code used. If you modify this file, you may extend this exception
to your version of the file, but you are not obligated to do so. If you
do not wish to do so, delete this exception statement from your version.
*/

// The necessary headers /////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <sched.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <asm/types.h>          /* for videodev2.h */
#include <linux/videodev2.h>

#ifndef NO_OVERLAYSNAP
#include "ovlmid.h"
#endif

#include "dh2226.h"
#include "s2226ioctl.h"			    // ioctl definitions for 2226 driver.


// Manifest constants ////////////////////////////////////////////////////////////////////////////////

#define ENABLED         1
#define DISABLED        0

#define TRUE            1
#define FALSE           0

#define MAXLEN          80              // Maximum length of user command string.
#define MAXNUMBOARDS    4               // Maximum number of boards supported by the driver.

// Stream mode:
#define MODE_IDLE                 0       // Halted.
#define MODE_PLAY_HD_1080I_50     1       // Streaming from file.
#define MODE_PLAY_HD_1080I_5994   2       // Streaming from file.
#define MODE_PLAY_HD_1080I_60     3       // Streaming from file.
#define MODE_PLAY_HD_720P_50      4       // Streaming from file.
#define MODE_PLAY_HD_720P_5994    5       // Streaming from file.
#define MODE_PLAY_HD_720P_60      6       // Streaming from file.
#define MODE_PLAY_SD_480I         7       // Streaming from file.
#define MODE_PLAY_SD_576I         8       // Streaming from file.
#define MODE_RECORD               9       // Streaming to file.

static int n_buffers = 0;

// Configuration constants ///////////////////////////////////////////////////////////////////////////

#define MAX_BOARDS	16		 // Maximum number of 2226 boards in the system.
#define MPEG_BUF_SIZE 8192 //4096		// Number of MPEG bytes to transfer in a single chunk.  Max size is 32K.
#define BLOCKING_MODE TRUE // Set to TRUE to use blocking read/write operations.

// Device name for the 2226 boards:
#define DEV_NAME		"/dev/video" //s2226v"


// Storage //////////////////////////////////////////////////////////////////

typedef struct BOARD {			// BOARD OBJECT -------
    int			handle;		    //  Handle to 2226 device.
    int			handle_decode;	//  Handle to 2226 decode device.
    int			mode;		//  Stream mode.
    int			stop_rq;	//  Stream Halt requested.
    char 		devname[MAXLEN];//  Device name for board.
    char 		fname[MAXLEN];	//  File name for mpeg data.
    pthread_t		io_thread_id;	//  Handle to stream i/o thread.
    unsigned int	mpeg_count;	//  Number of MPEG bytes transferred.
    int			flags;		//  Status flags from i/o thread to main thread.
    struct {				//  Settings for the board
	int		sel_video_input; // selected video input
	int		cur_video_input; // current board input(required if decode)
    int     valid_input;     // set to 1 if input set was valid, 0 otherwise
        int audio_bitrate;
/*
        mode_param_t	video_bitrate;
	mode_param_t	audio_mode;
	mode_param_t	audio_stereo;
	// audio analog parameters
    mode_param_t	audio_route; // input gain
    mode_param_t	audio_gain; // input gain
    mode_param_t	audio_agcgain; // AGC gain
	mode_param_t	audio_agc_on;
    mode_param_t	audio_balanced; // 0 unbalanced, 1 balanced
*/
    } settings;
} BOARD;

static int vid_status(struct BOARD *b, int *lock);

struct buffer {
	void *start;
	size_t length;
};

struct buffer *buffers = NULL;


static struct BOARD bd[MAX_BOARDS];	// Board objects.
static int num_boards;                  // Actual number of boards.
static int act_bdnum	= 0;            // Board 0 is active by default.


// Function prototypes ///////////////////////////////////////////////////////////////////////////////

static int app_init( void );
static void app_shutdown( void );
static void cmd_processor( void );
static void *io_thread_func( void* arg );
static void get_attributes( struct BOARD *b );
static void start_stream( struct BOARD *b, char *fname, int new_stream_mode );
static void stop_stream( struct BOARD *b );
static void prompt( void );
static int change_input(BOARD *b);


///////////////////////////////////////////////////////////////////////////
// Init the system and execute user commands until termination.

int main( int argc, char *argv[] )
{
    int rtnval;

    // Attempt to open and initialize all boards.
    rtnval = app_init();

    // If any board successfully opened and initialized ...
    if ( rtnval > 0 )
    {
    // Execute user commands until Application Quit command is invoked.
    cmd_processor();
    } else {
	printf("\nno boards found");
    }

    // Close all opened boards.
    app_shutdown();

    // Exit application.
    return 0;
}

///////////////////////////////////////////////////////////////////////
// Ceck board type.  Returns: 1 if 2226 encode, 2 if decode, 0 in no board or any other code.
static int check_type(char *dev_name)
{
    int handle;
	struct v4l2_capability cap;
    int rc;
    handle = open( dev_name, O_RDWR );
    if (handle == -1) {
        return( 0 );
    }
	rc = ioctl(handle, VIDIOC_QUERYCAP, &cap);
    if (rc != 0) {
        close(handle);
        return 0;
    }
	fprintf(stderr, "Card: %s\n", cap.card);
	if (strncmp((const char *)cap.card, "2226 Decode", 11)) {
        if (strncmp((const char *) cap.card, "2226 H.264", 10)) {
            if (strncmp((const char *)cap.card, "2226 Preview", 12))
                fprintf (stderr, "Sorry, \"%s\" is not a Sensoray 2226 device\n",
                         dev_name);
            return 0;
        } else {
            close(handle);
            return 1;
        }
    } else {
        close(handle);
        return 2;
    }
}



static void init_mmap(int fd)
{
	struct v4l2_requestbuffers req;
    memset(&req, 0, sizeof(req));
	req.count               = 8;
	req.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	req.memory              = V4L2_MEMORY_MMAP;

	if (-1 == ioctl (fd, VIDIOC_REQBUFS, &req)) {
		if (EINVAL == errno) {
			fprintf (stderr, "device does not support "
                     "memory mapping\n");
			return;
		} else {
			printf("VIDIOC_REQBUFS failed\n");
            return;
		}
	}

	if (req.count < 2) {
        fprintf (stderr, "device does not support "
                 "memory mapping\n");
        return;

	}

	buffers = calloc (req.count, sizeof (*buffers));

	if (!buffers) {
		fprintf (stderr, "Out of memory\n");
		exit (EXIT_FAILURE);
	}

	for (n_buffers = 0; n_buffers < req.count; ++n_buffers) {
		struct v4l2_buffer buf;
        memset(&buf, 0, sizeof(buf));
		buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory      = V4L2_MEMORY_MMAP;
		buf.index       = n_buffers;

		if (-1 == ioctl (fd, VIDIOC_QUERYBUF, &buf)) {
			printf("VIDIOC_QUERYBUF error");
            return;
        }
		
		//fprintf(stderr, "buf.length = %d\n", buf.length);

		buffers[n_buffers].length = buf.length;
		buffers[n_buffers].start =
			mmap (NULL /* start anywhere */,
			      buf.length,
			      PROT_READ | PROT_WRITE /* required */,
			      MAP_SHARED /* recommended */,
			      fd, buf.m.offset);

		if (MAP_FAILED == buffers[n_buffers].start) {
			printf("mmap failed\n");
            return;
        }
	}
}

///////////////////////////////////////////////////////////////////////
// Open and initialize one board.  Returns: 0=success, -1=error.

static int board_open( int board_num, char *dev_name )
{
    // Set pointer to board's object.
    struct BOARD *b = &bd[board_num];
    struct v4l2_control ctrl;
    // Initialize object.
    b->mode	= MODE_IDLE;
    b->fname[0]	= 0;
    b->flags	= 0;

    // Open device.
    b->handle = open( dev_name, O_RDWR | (BLOCKING_MODE ? 0 : O_NONBLOCK));
    if ( b->handle == -1 )
    {
        printf( "Driver open error\n" );
        return( -1 );
    }
    init_mmap(b->handle);
    strcpy( b->devname, dev_name );

    // Program the default settings.
/*
    b->settings.sel_video_input = INPUT_COMP0_480I;
    b->settings.valid_input = 0;
    b->settings.video_bitrate.idx = S2226_IDX_VBITRATE;
    b->settings.video_bitrate.val = 2300;
    b->settings.audio_mode.idx = S2226_IDX_AMODE;
    b->settings.audio_mode.val = 0; //AMODE_MP1L2;
    b->settings.audio_bitrate.idx = S2226_IDX_ABITRATE;

    b->settings.audio_stereo.idx = S2226_IDX_STEREO;
    b->settings.audio_stereo.val = 1;

    b->settings.audio_gain.idx = S2226_IDX_AUDIN_GAIN;
    b->settings.audio_gain.val = 0;

    b->settings.audio_agcgain.idx = S2226_IDX_AUDIN_AGC_GAIN;
    b->settings.audio_agcgain.val = 0;

    b->settings.audio_agc_on.idx = S2226_IDX_AUDIN_AGC_ON;
    b->settings.audio_agc_on.val = 0;

    b->settings.audio_route.idx = S2226_IDX_AUDROUTE;
    b->settings.audio_route.val = S2226_AUDIOROUTE_LINE1L;

    b->settings.audio_balanced.idx = S2226_IDX_AUDIN_BAL;
    b->settings.audio_balanced.val = 0;

    b->settings.cur_video_input = -1;

    ioctl(b->handle, S2226_VIDIOC_SET_MODE, &b->settings.video_bitrate);
    ioctl(b->handle, S2226_VIDIOC_SET_MODE, &b->settings.audio_mode);
    ioctl(b->handle, S2226_VIDIOC_SET_MODE, &b->settings.audio_bitrate);
    ioctl(b->handle, S2226_VIDIOC_SET_MODE, &b->settings.audio_stereo);
*/

    ctrl.id = V4L2_CID_MPEG_AUDIO_L2_BITRATE;
    ctrl.value = V4L2_MPEG_AUDIO_L2_BITRATE_192K;
    b->settings.audio_bitrate = ctrl.value;
    ioctl(b->handle, VIDIOC_S_CTRL, &ctrl);

    ctrl.id = V4L2_CID_MPEG_AUDIO_MODE;
    ctrl.value = V4L2_MPEG_AUDIO_MODE_STEREO;
    ioctl(b->handle, VIDIOC_S_CTRL, &ctrl);

    // set input for faster initial startup if input unchanged.
    if (change_input(b) != 0) {
        b->settings.valid_input = 0;
    } else
        b->settings.valid_input = 1;

#ifndef NO_OVERLAYSNAP
    // clear overlay
    S2226_ClearOverlay(b->handle);
    S2226_UpdateOverlayNoInput(b->handle);
#endif
    return 0;
}



static int board_open_decode( int board_num, char *dev_name )
{
    // Set pointer to board's object.
    struct BOARD *b = &bd[board_num];
    b->handle_decode = open( dev_name, O_RDWR | (BLOCKING_MODE ? 0 : O_NONBLOCK));
    printf("decode handle %d %s\n", b->handle_decode, dev_name);
    ioctl(b->handle_decode, 0, 0);
    ioctl(b->handle_decode, 0, 0);
    ioctl(b->handle_decode, 0, 0);
    if ( b->handle_decode == -1 )
    {
        printf( "Driver open error\n" );
        return( -1 );
    }
    return 0;
}


///////////////////////////////////////////////////////////////////////
// Open and initialize all boards.  Returns: number of boards.

static int app_init( void )
{
    int		i;
    char	dev_name[80];
    int type;

    // For each board ...
    for ( i = num_boards = 0; i < MAX_BOARDS; i++ ) {
        // Specify board's device name.
        sprintf( dev_name, DEV_NAME "%d", i );
        // Check board type.  Go to next if not sx16.
        type = check_type( dev_name );
        if ((type != 1) && (type != 2)) {
            if( type != 0 )
                printf( "Not a 2226 board #%d found as %s\n", num_boards, dev_name );
            continue;
        }
        // Open the board. Go to next if fail.
        if (type == 1) {
            if (board_open( num_boards, dev_name))
                continue;
        } else if( board_open_decode( num_boards, dev_name ) )
            continue;
        printf( "%d board #%d found as %s\n", type, num_boards, dev_name );
        if( num_boards == MAX_BOARDS - 1 )
            break;
        if (type == 2)
            num_boards++;
    }
    return num_boards;
}

//////////////////////////////////////////////////////////////
// Terminate all stream threads and close all boards.

static void app_shutdown( void )
{
    struct BOARD *b;
    int i;
    // For each board ...
    for (b = bd; b < &bd[num_boards]; b++) {
	// We are finished if this board was never opened.
	if ( b->handle == -1 )
	    break;

	// Terminate the board's I/O thread if it exists.
	stop_stream( b );

	// Close the board.
	for (i = 0; i < n_buffers; ++i)
	  if (-1 == munmap(buffers[i].start, buffers[i].length))
            printf("unmap fail\n");
	free (buffers);
    close( b->handle );
    }

    printf( "\n" );
}

//////////////////////////////////////////////////////////////
// Display the user prompt.

static void prompt( void )
{
    printf( "2226demo[%d]>", act_bdnum );
    fflush( stdout );
}

//////////////////////////////////////////////////////////////
// Display the command menu.

static void show_menu( void )
{
    printf( "+-2226DEMO-COMMANDS------------------------------------------------------------+\n" );
    printf( "|   board n ....... select board: 0..n                                         |\n" );
    printf( "|   get ........... display current settings                                   |\n" );
    printf( "|   h[elp] ........ display this list                                          |\n" );
    printf( "|   q[uit] ........ quit program                                               |\n" );
    printf( "+-AUDIO------------------------------------------------------------------------+\n" );
    printf( "|   arate n ....... set kbits/second:                                          |\n" );
    printf( "|                     default=192 (other values 224, 256)                      |\n" );
    printf( "|   aformat n ..... set format:                                                |\n" );
    printf( "|                     0=MPEG-1 layer 2 (default)                               |\n" );
    printf( "|   stereo n ...... set stereo mode: 0=mono, 1=stereo                          |\n" );
    printf( "|   aroute n ...... set audio routing(0 line1, 1 line2, 2 line1 passthrough)   |\n" );
    printf( "|   gain n ........ set audio input gain(0-default, max 118 (steps 0.5 db))    |\n" );
    printf( "|   agcgain n ..... set audio input AGC gain(0-default, max 118 (steps 0.5 db))|\n" );
    printf( "|   agc n ......... turn on/off audio input AGC(0-default off, 1-on)           |\n" );
    printf( "|   bal n ......... set audio input as balanced(0-off, 1-balanced)             |\n" );
    printf( "+-VIDEO------------------------------------------------------------------------+\n" );
    printf( "|   vstat    ...... query video status:                                        |\n" );
    printf( "|   vinput n ...... set video input:                                           |\n" );
    printf( "|                     0=CV0 720x480 (29.97i) ITU-R BT.656-4                    |\n" );
    printf( "|                     1=CV0 720x576 (25i) ITU-R BT.656-4                       |\n" );
    printf( "|                     2=SV0 720x480 (29.97i) ITU-R BT.656-4                    |\n" );
    printf( "|                     3=SV0 720x576 (25i) ITU-R BT.656-4                       |\n" );
    printf( "|                     4=SDI 720x480 (29.97i) ITU-R BT.656-4                    |\n" );
    printf( "|                     5=SDI 720x480 COLORBARS (29.97i) ITU-R BT.656-4          |\n" );
    printf( "|                     6=SDI 720x576 (25i) ITU-R BT.656-4                       |\n" );
    printf( "|                     7=SDI 720x576 COLORBARS (25i) ITU-R BT.656-4             |\n" );
    printf( "|                     8=SDI 1920x1080 (50i) SMPTE274M                          |\n" );
    printf( "|                     9=SDI 1920x1080 (59.94i) SMPTE274M                       |\n" );
    printf( "|                    10=SDI 1920x1080 (60i) SMPTE274M                          |\n" );
    printf( "|                    11=SDI 1920x1080 COLORBARS (60i) SMPTE274M                |\n" );
    printf( "|                    12=SDI 1280x720 (50p) SMPTE296M-2001                      |\n" );
    printf( "|                    13=SDI 1280x720 (59.94p) SMPTE296M-2001                   |\n" );
    printf( "|                    14=SDI 1280x720 (60p) SMPTE296M-2001                      |\n" );
    printf( "|                    40=SDI 1920x1080 (24p) SMPTE274M                          |\n" );
    printf( "|                    41=SDI 1920x1080 (23.98p) SMPTE274M                       |\n" );
    printf( "|                    15=SDI 1280x720 COLORBARS (60p) SMPTE296M-2001            |\n" );
    printf( "|                    23=SDI 1920x1080 COLORBARS (50i) SMPTE274M                |\n" );
    printf( "|                    29=SDI 1920x1080 COLORBARS (59.94i) SMPTE274M             |\n" );
    printf( "|                    30=SDI 1280x720 COLORBARS (59.94i) SMPTE274M              |\n" );
    printf( "|                    31=SDI 1280x720 COLORBARS (50p) SMPTE274M                 |\n" );
    printf( "|                    34=SDI 1920x1080 COLORBARS (24p) SMPTE274M                |\n" );
    printf( "|                    35=SDI 1920x1080 COLORBARS (23.98p) SMPTE274M             |\n" );
    printf( "|   bright n ...... set luminance brightness: 0..255 (default 128)             |\n" );
    printf( "|   contrast n .... set luminance contrast: 0..127 (default 64)                |\n" );
    printf( "|   vsat n ........ set chrominance saturation: 0..127 (default 64)            |\n" );
    printf( "|   vhue n ........ set video hue: -64..64 (default 0)                         |\n" );
    printf( "+-MPEG-------------------------------------------------------------------------+\n" );
    printf( "|   vrate n ....... set video kbits/second: 1000..19000                        |\n" );
    printf( "+-STREAM-----------------------------------------------------------------------+\n" );
    printf( "|   rec fname ..... record to MPEG file; default name is \"tmp.mpg\"             |\n" );
    printf( "|   p480i fname..play 480i content from MPEG file; dflt name is \"tmp.mpg\"      |\n" );
    printf( "|   p576i fname..play 480i content from MPEG file; dflt name is \"tmp.mpg\"      |\n" );
    printf( "|   p720p_50 fname...play HD 720p 50Hz content from MPEG file;                 |\n" );
    printf( "|                     default name is \"tmp.mpg\"                                |\n" );
    printf( "|   p720p_5994 fname..play HD 720p 59.94Hz content from MPEG file;             |\n" );
    printf( "|                     default name is \"tmp.mpg\"                                |\n" );
    printf( "|   p720p_60 fname... play HD 720p 60Hz content from MPEG file;                |\n" );
    printf( "|                     default name is \"tmp.mpg\"                                |\n" );
    printf( "|   p1080i_50 fname... play HD 1080i 50Hz content from MPEG file;              |\n" );
    printf( "|                     default name is \"tmp.mpg\"                                |\n" );
    printf( "|   p1080i_5994 fname... play HD 1080i 59.94Hz content from MPEG file;         |\n" );
    printf( "|                     default name is \"tmp.mpg\"                                |\n" );
    printf( "|   p1080i_60 fname... play HD 1080i 60Hz content from MPEG file;              |\n" );
    printf( "|                     default name is \"tmp.mpg\"                                |\n" );
    printf( "|   stop .......... stop MPEG stream and close file                            |\n" );
    printf( "+-Snapshot---------------------------------------------------------------------+\n" );
    printf( "+Each snapshot method below uses a different field interpolation type.         |\n");
    printf( "+snap0 is no field interpolation, 1 is duplicate, 2 is merge, 3 is interpolate |\n");
    printf( "+bmp extension will be added to filename.                                      |\n");
    printf( "|   snap0 filename...... save uncompressed image from hardware to BMP file     |\n" );
    printf( "|   snap1 filename...... save uncompressed image from hardware to BMP file     |\n" );
    printf( "|   snap2 filename...... save uncompressed image from hardware to BMP file     |\n" );
    printf( "|   snap3 filename...... save uncompressed image from hardware to BMP file     |\n" );
#ifndef NO_OVERLAYSNAP
    printf( "+-Overlay----------------------------------------------------------------------+\n" );
    printf( "|Note, by default overlay added to all regions(H264 stream, snapshots, vid out)|\n");
    printf( "|   text i x y captext .updates overlay window i(0-7) with text                |\n" );
    printf( "|   text2 i x y captext .updates overlay window i(0-7) with transparent text   |\n" );
    printf( "|   bitmap i x y bmpfile.updates overlay window i(0-7) with bmp (eg. logo.bmp) |\n" );
    printf( "|   move i x y..........moves window i to new position                         |\n" );
    printf( "|   del  i..............deletes window i                                       |\n" );
    printf( "|   clear ..............clears all overlays                                    |\n" );
#endif
    printf( "+------------------------------------------------------------------------------+\n" );
}

///////////////////////////////////////////////////////////////////////////////
// Decode and execute user command.  Returns TRUE if Quit selected.

static int exec_command( char *cmd, char *arg, char *arg2, char *arg3, char *arg4, char *arg5 )
{
    struct BOARD	*b;

    // Convert argument to integer value (not relevant for all commands).
    int arg_val = -1;
    sscanf( arg, "%d", &arg_val);
    // Decode and execute command -----------------------------------

    // Terminate application.
    if ( cmd[0] == 'q' )
        return TRUE;

    // Select board.
    if ( strcmp( cmd, "board" ) == 0 )
    {
	if ( ( arg_val >= 0 ) && ( arg_val < num_boards ) )
	    act_bdnum = arg_val;
	else
	    printf( "Illegal or missing argument; must be in range 0..%d\n", num_boards - 1 );
	prompt();
	return FALSE;
    }

    // Show help.
    if ( ( cmd[0] == 'h' ) || ( cmd[0] == '?' ) )
    {
	show_menu();
	prompt();
	return FALSE;
    }

    // Prompt for user command.
    prompt();

    // Set pointer to active board's object.
    b = &bd[act_bdnum];
    if (!strcmp(cmd, "arate")) {
	if (arg_val >= 192 && arg_val <= 256) {
        struct v4l2_control ctrl;
        ctrl.id = V4L2_CID_MPEG_AUDIO_L2_BITRATE;
        switch (arg_val) {
        case 192:
        default:
            ctrl.value = V4L2_MPEG_AUDIO_L2_BITRATE_192K;
            break;
        case 224:
            ctrl.value = V4L2_MPEG_AUDIO_L2_BITRATE_224K;
            break;
        case 256:
            ctrl.value = V4L2_MPEG_AUDIO_L2_BITRATE_256K;
            break;
        }
        b->settings.audio_bitrate = ctrl.value;
	    ioctl( b->handle, VIDIOC_S_CTRL, &ctrl);
	} else
	    printf("invalid setting for arate\n"); }
    else if (!strcmp(cmd,"aformat")) {
	if (arg_val ==0) {
        // TODO
	    //b->settings.audio_mode.val = arg_val;
	    //ioctl( b->handle, S2226_VIDIOC_SET_MODE, &b->settings.audio_mode);
	} else
	    printf("invalid setting for aformat\n"); }
    else if (!strcmp(cmd,"vstat")) {
      int rc;
      int lock;
      rc = vid_status(b, &lock);
      if (rc != 0)
	printf("failed to query status\n");
      else {
	if (lock)
	  printf("Video valid and locked\n");
	else
	  printf("No video lock\n");
      }
    } else if (!strcmp(cmd,"vinput")) {
        if ((arg_val >= 0 && arg_val <= 15) || (arg_val == 23) 
            || (arg_val >= 40 && arg_val <= 41) 
            || (arg_val >= 34 && arg_val <= 35) 
            || ((arg_val <= 31) && (arg_val >= 29))) {
	    b->settings.sel_video_input = arg_val;

	    if (change_input(b) != 0) {
		    printf("Input not present or busy. Input must be valid before recording stream\n");
		    b->settings.valid_input = 0;
	    } else
		    b->settings.valid_input = 1;
	} else
	    printf("invalid setting for vinput\n"); }
    else if (!strcmp(cmd,"stereo")) {
	if (arg_val >= 0 && arg_val <= 1) {
        struct v4l2_control ctrl;
        ctrl.id = V4L2_CID_MPEG_AUDIO_MODE;
        ctrl.value = (arg_val) ? V4L2_MPEG_AUDIO_MODE_STEREO : V4L2_MPEG_AUDIO_MODE_MONO;
        ioctl(b->handle, VIDIOC_S_CTRL, &ctrl);
	} else
	    printf("invalid setting for stereo\n"); }
/*
    else if (!strcmp(cmd,"route")) {
	if (arg_val >= 0 && arg_val <= 3) {
	    b->settings.audio_route.val = arg_val;
	    ioctl( b->handle, S2226_VIDIOC_SET_MODE, &b->settings.audio_route);
	} else
	    printf("invalid setting for audio route\n"); }
*/
    else if (!strcmp(cmd,"agc")) {
	if (arg_val >= 0 && arg_val <= 1) {
        struct v4l2_control ctrl;
        ctrl.id = S2226_CID_AUDIN_AGC_ON_R;
        ctrl.value = arg_val;
        ioctl(b->handle, VIDIOC_S_CTRL, &ctrl);
        ctrl.id = S2226_CID_AUDIN_AGC_ON_L;
        ctrl.value = arg_val;
        ioctl(b->handle, VIDIOC_S_CTRL, &ctrl);
	} else
	    printf("invalid setting for AGC on/off\n"); }
    else if (!strcmp(cmd,"bal")) {
	if (arg_val >= 0 && arg_val <= 1) {
        struct v4l2_control ctrl;
        ctrl.id = S2226_CID_AUDIN_BAL_R;
        ctrl.value = arg_val;
        ioctl(b->handle, VIDIOC_S_CTRL, &ctrl);
        ctrl.id = S2226_CID_AUDIN_BAL_L;
        ctrl.value = arg_val;
        ioctl(b->handle, VIDIOC_S_CTRL, &ctrl);
	} else
	    printf("invalid setting for balanced on/off\n"); }
    else if (!strcmp(cmd,"agcgain")) {
	if (arg_val >= 0 && arg_val <= 118) {
        struct v4l2_control ctrl;
        ctrl.id = S2226_CID_AUDIN_AGC_GAIN_R;
        ctrl.value = arg_val;
        ioctl(b->handle, VIDIOC_S_CTRL, &ctrl);
        ctrl.id = S2226_CID_AUDIN_AGC_GAIN_L;
        ctrl.value = arg_val;
        ioctl(b->handle, VIDIOC_S_CTRL, &ctrl);
	} else
	    printf("invalid setting for AGC gain\n"); }
    else if (!strcmp(cmd,"gain")) {
	if (arg_val >= 0 && arg_val <= 118) {
        struct v4l2_control ctrl;
        ctrl.id = S2226_CID_AUDIN_GAIN_R;
        ctrl.value = arg_val;
        ioctl(b->handle, VIDIOC_S_CTRL, &ctrl);
        ctrl.id = S2226_CID_AUDIN_GAIN_L;
        ctrl.value = arg_val;
        ioctl(b->handle, VIDIOC_S_CTRL, &ctrl);
	} else
	    printf("invalid setting for audio gain\n"); }
    else if (!strcmp(cmd,"vrate")) {
	if (arg_val >= 1000 && arg_val <= 19000) {
        struct v4l2_control ctrl;
        ctrl.id = V4L2_CID_MPEG_VIDEO_BITRATE;
        ctrl.value = arg_val * 1000;
	    ioctl(b->handle, VIDIOC_S_CTRL, &ctrl);
	} else
	    printf("invalid setting for vrate\n"); }
  
    else if (!strcmp(cmd,"bright")) {
	if (arg_val >= 0 && arg_val <= 255) {
        struct v4l2_control ctrl;
        ctrl.id = V4L2_CID_BRIGHTNESS;
        ctrl.value = arg_val;
		ioctl( b->handle, VIDIOC_S_CTRL, &ctrl);
	} else
	    printf("invalid setting for level\n"); }
    else if (!strcmp(cmd,"contrast")) {
	if (arg_val >= 0 && arg_val <= 127) {
        struct v4l2_control ctrl;
        ctrl.id = V4L2_CID_CONTRAST;
        ctrl.value = arg_val;
		ioctl( b->handle, VIDIOC_S_CTRL, &ctrl);
	} else
	    printf("invalid setting for level\n"); }
    else if (!strcmp(cmd,"vsat")) {
	if (arg_val >= 0 && arg_val <= 127) {
        struct v4l2_control ctrl;
        ctrl.id = V4L2_CID_SATURATION;
        ctrl.value = arg_val;
		ioctl( b->handle, VIDIOC_S_CTRL, &ctrl);
	} else
	    printf("invalid setting for level\n"); }
    else if (!strcmp(cmd,"vhue")) { 
	if (arg_val >= -64 && arg_val <= 64) {
        struct v4l2_control ctrl;
        ctrl.id = V4L2_CID_HUE;
        ctrl.value = arg_val;
		ioctl( b->handle, VIDIOC_S_CTRL, &ctrl);
	} else
	    printf("invalid setting for level\n"); }
#ifndef NO_OVERLAYSNAP
    else if (!strcmp(cmd,"snap0")) { 
        char *file = arg;
        if (!b->settings.valid_input) {
            printf("input not valid or not set. Use vinput command.\n");
            prompt();
        } else {
            S2226_SetMergeMethod(MID2226_FIELDALG_NONE, b->handle);
            S2226_SnapshotToFile(file, MID2226_FILE_BMP, 1, 1, b->handle); }}
    else if (!strcmp(cmd,"snap1")) { 
        char *file = arg;
        if (!b->settings.valid_input) {
            printf("input not valid or not set. Use vinput command.\n");
            prompt();
        } else {
            S2226_SetMergeMethod(MID2226_FIELDALG_DUP, b->handle);
            S2226_SnapshotToFile(file, MID2226_FILE_BMP, 0, 0, b->handle); }}
    else if (!strcmp(cmd,"snap2")) { 
        char *file = arg;
        if (!b->settings.valid_input) {
            printf("input not valid or not set. Use vinput command\n");
            prompt();
        } else {
            S2226_SetMergeMethod(MID2226_FIELDALG_MERGE, b->handle);
            S2226_SnapshotToFile(file, MID2226_FILE_BMP, 0, 0, b->handle); }}
    else if (!strcmp(cmd,"snap3")) { 
        char *file = arg;
        if (!b->settings.valid_input) {
            printf("input not valid or not set. Use vinput command\n");
            prompt();
        } else {
            S2226_SetMergeMethod(MID2226_FIELDALG_INTER, b->handle);
            S2226_SnapshotToFile(file, MID2226_FILE_BMP, 0, 0, b->handle); }}
    else if (!strcmp(cmd,"text")) { 
        int x, y;
        char *text = arg4;
        sscanf(arg2, "%d", &x);
        sscanf(arg3, "%d", &y);
        if (arg_val >= 0 && arg_val <= 7) {
            overlay_text_t ovl;
            int rc;
            memset(&ovl, 0, sizeof(ovl));
            if (text == NULL)
                strcpy(ovl.text, "Caption text");
            else
                strcpy(ovl.text, text);
            ovl.fontsize = 30;
            // font file can be change to any true type font (TTF)
            //strcpy(ovl.fontfile, "arial.ttf");
            strcpy(ovl.fontfile, "DejaVuSans.ttf");
            ovl.txtcolor = RGB(255,255,255);
            ovl.backcolor = RGB(5,5,5); //transparent is 0,0,0
            rc = S2226_OverlayTextIdx(arg_val, x, y, &ovl, 0xff, b->handle);
            if (rc != 0) {
                printf("S2226_OverlayTextIdx returned %d\n", rc);
            }
            rc = S2226_UpdateOverlayNoInput(b->handle);
            if (rc != 0) {
                printf("S2226_UpdateOverlayNoInput returned %d\n", rc);
            }
        } else
            printf("invalid setting for index\n"); }
    else if (!strcmp(cmd,"text2")) { 
        int x, y;

        char *text = arg4;
        sscanf(arg2, "%d", &x);
        sscanf(arg3, "%d", &y);
        if (arg_val >= 0 && arg_val <= 7) {
            overlay_text_t ovl;
            memset(&ovl, 0, sizeof(ovl));
            if (text == NULL)
                strcpy(ovl.text, "Caption text");
            else
                strcpy(ovl.text, text);
            ovl.fontsize = 30;
            strcpy(ovl.fontfile, "DejaVuSans.ttf");
            ovl.txtcolor = RGB(255,254,255);
            ovl.backcolor = RGB(0,0,0); //transparent is 0,0,0
            S2226_OverlayTextIdx(arg_val, x, y, &ovl, 0xff, b->handle);
            S2226_UpdateOverlayNoInput(b->handle);
        } else
            printf("invalid setting for index\n"); }
    else if (!strcmp(cmd,"bitmap")) { 
        int x, y;
        char *file = arg4;
        sscanf(arg2, "%d", &x);
        sscanf(arg3, "%d", &y);
        if (arg_val >= 0 && arg_val <= 7) {
            S2226_OverlayImageIdx(arg_val, x, y, file, 0xff, b->handle);
            S2226_UpdateOverlayNoInput(b->handle);
        } else
            printf("invalid setting for index\n"); }
    else if (!strcmp(cmd,"move")) { 
        int x, y;
        sscanf(arg2, "%d", &x);
        sscanf(arg3, "%d", &y);
        if (arg_val >= 0 && arg_val <= 7) {
            S2226_MoveOverlay(arg_val, x, y, b->handle);
            S2226_UpdateOverlayNoInput(b->handle);
        } else
            printf("invalid setting for index\n"); }
    else if (!strcmp(cmd,"del")) { 
        if (arg_val >= 0 && arg_val <= 7) {
            S2226_OverlayDel(arg_val, b->handle);
            S2226_UpdateOverlayNoInput(b->handle);
        } else
            printf("invalid setting for index\n"); }
    else if (!strcmp(cmd, "clear")) { 
        S2226_ClearOverlay(b->handle);
        S2226_UpdateOverlayNoInput(b->handle);
    }
#endif
	// Display current settings.
    else if ( strcmp( cmd, "get"       ) == 0 )	{
	    get_attributes( b ); prompt(); }
    // Control MPEG stream mode.
    else if ( strcmp( cmd, "rec"       ) == 0 )
	    start_stream( b, ( arg[0] == 0 ) ? "tmp.mpg" : arg, MODE_RECORD );
    else if ( strcmp( cmd, "p1080i_50"    ) == 0 )
	    start_stream( b, ( arg[0] == 0 ) ? "tmp.mpg" : arg, MODE_PLAY_HD_1080I_50);
    else if ( strcmp( cmd, "p1080i_60"    ) == 0 )
	    start_stream( b, ( arg[0] == 0 ) ? "tmp.mpg" : arg, MODE_PLAY_HD_1080I_60);
    else if ( strcmp( cmd, "p1080i_5994"    ) == 0 )
	    start_stream( b, ( arg[0] == 0 ) ? "tmp.mpg" : arg, MODE_PLAY_HD_1080I_5994);
    else if ( strcmp( cmd, "p720p_60"     ) == 0 )
	    start_stream( b, ( arg[0] == 0 ) ? "tmp.mpg" : arg, MODE_PLAY_HD_720P_60);
    else if ( strcmp( cmd, "p720p_5994"     ) == 0 )
	    start_stream( b, ( arg[0] == 0 ) ? "tmp.mpg" : arg, MODE_PLAY_HD_720P_5994);
    else if ( strcmp( cmd, "p720p_50"     ) == 0 )
	    start_stream( b, ( arg[0] == 0 ) ? "tmp.mpg" : arg, MODE_PLAY_HD_720P_50);
    else if ( strcmp( cmd, "p480i"     ) == 0 )
	    start_stream( b, ( arg[0] == 0 ) ? "tmp.mpg" : arg, MODE_PLAY_SD_480I);
    else if ( strcmp( cmd, "p576i"     ) == 0 )
	    start_stream( b, ( arg[0] == 0 ) ? "tmp.mpg" : arg, MODE_PLAY_SD_576I);
    else if ( strcmp( cmd, "stop"      ) == 0 )
	    stop_stream( b );
	// Illegal command.
    else if ( cmd[0] != 0 )
    {
	printf( "Invalid command: %s\n", cmd );
	prompt();
    }
				   
    return FALSE;
}

////////////////////////////////////////////////////////////////////
// Execute user commands until Application Quit is requested.

static void cmd_processor( void )
{
    char	cmdbuf[MAXLEN];
    char	cmd[MAXLEN];
    char	arg[MAXLEN];
    char	arg2[MAXLEN];
    char	arg3[MAXLEN];
    char	arg4[MAXLEN];
    char	arg5[MAXLEN];
    char	*c;

    // Display the command menu.
    show_menu();

    // Prompt for user command.
    prompt();

    // Repeat until Application Terminate is requested ...
    for ( ; ; )
    {
	// Input user command string, and if no errors encountered ...
	if ( fgets( cmdbuf, MAXLEN - 1, stdin ) != NULL )
	{
	    // Truncate newline char at end of command string.
	    cmdbuf[ strlen( cmdbuf ) - 1 ] = 0;

	    // If command line is not blank ...
	    if ( cmdbuf[0] != 0 )
	    {
	        // Force command string to lower case, except command=="text".
		if ( strncmp( cmdbuf, "text", 4 ) != 0 )
		    for ( c = cmdbuf; *c != 0; c++ )
			*c = tolower( *c );

	    // Parse command string into opcode and argument.
	    cmd[0] = arg[0] = arg2[0] = arg3[0] = arg4[0] = arg5[0] = 0;
	    sscanf( cmdbuf, "%s %s %s %s %s %s", cmd, arg, arg2, arg3, arg4, arg5);

	    // Decode and execute command
	    if ( exec_command( cmd, arg, arg2, arg3, arg4, arg5 ) )
		break;
	    }
	    else
	    {
		// Prompt for user command.
		prompt();
	    }
	}
    }
}




/////////////////////////////////////////////////////
// Fetch and display board's current settings.


static char *txt_vinput[] = {
"CV0 720x480 (29.97i) ITU-R BT.656-4",
"CV0 720x576 (25i) ITU-R BT.656-4",
"SV0 720x480 (29.97i) ITU-R BT.656-4",
"SV0 720x576 (25i) ITU-R BT.656-4",
"SDI 720x480 (29.97i) ITU-R BT.656-4",
"SDI 720x480 COLORBARS (29.97i) ITU-R BT.656-4",
"SDI 720x576 (25i) ITU-R BT.656-4",
"SDI 720x576 COLORBARS (25i) ITU-R BT.656-4",
"SDI 1920x1080 (50i) SMPTE274M",
"SDI 1920x1080 (59.94i) SMPTE274M",
"SDI 1920x1080 (60i) SMPTE274M",
"SDI 1920x1080 COLORBARS (60i) SMPTE274M",
"SDI 1280x720 (50p) SMPTE296M-2001",
"SDI 1280x720 (59.94p) SMPTE296M-2001",
"SDI 1280x720 (60p) SMPTE296M-2001",
"SDI 1280x720 COLORBARS (60p) SMPTE296M-2001",
};
						   
// Fetch and display the attributes.
static void get_attributes( struct BOARD *b )
{
    struct io_reg cmd;
    int rc;
    printf( "\nCURRENT SETTINGS -----------\n" );
    printf( "video format         = %s\n", txt_vinput[b->settings.sel_video_input]);
    //printf( "video bitrate        = %d\n", b->settings.video_bitrate.val);
//    printf( "stereo               = %d\n", b->settings.audio_stereo.val);
//    printf( "audio bitrate        = %d\n", b->settings.audio_bitrate.val);
    cmd.addr = ATTR_ARM_FW_VERSION;
    rc = ioctl( b->handle, S2226_VIDIOC_GET_ATTR, &cmd);
    if (rc != 0) {
	    printf("VIDIOC_GET_ATTR failed, %x\n", rc);
	    return;
    }
    printf( "ARM firmware version = 0x%04x\n", cmd.val);
    cmd.addr = (0x21 << 1); ioctl( b->handle, S2226_VIDIOC_FPGA_RD, &cmd);
    printf( "FPGA firmware version = %d\n", cmd.val & 0x7ff);
    printf( "board id              = %d\n", cmd.val >> 11);
}


#define VIDSTAT_REG 0x04
static int vid_status(struct BOARD *b, int *lock)
{
    struct io_reg cmd;
    int rc;
    cmd.addr = (0x4 << 1);
    rc = ioctl(b->handle, S2226_VIDIOC_FPGA_RD, &cmd);
    if (rc != 0)
      return rc;
    *lock = 0;
    if (cmd.val & 0x1)
      *lock = 1;
    if (cmd.val & 0x80)
      *lock = 0;
    return rc;
}


///////////////////////////////////////////////////
// Start MPEG data streaming to/from a file.

static void start_stream( struct BOARD *b, char *fname, int new_stream_mode )
{
    pthread_attr_t attr;
    int lock;
    int rc;

    rc = vid_status(b, &lock);

    if (rc != 0) {
      printf("failed to get video status\n");
      return;
    }
    if (lock != 1) {
      printf("no video lock, stream not started\n");
      return;
    }
    
    
    
    // If we are already streaming ...
    if ( b->mode != MODE_IDLE )
    {
	// Abort command.
	printf( "already streaming\n" );
	prompt();
	return;
    }

    // Make the MPEG filename visible to the I/O thread.
    strcpy( b->fname, fname );

    // Indicate the new stream mode.
    b->mode = new_stream_mode;

    // Clear the stream thread's termination request.
    b->stop_rq = 0;

    // Create detached thread that will manage the MPEG stream flow.
    pthread_attr_init( &attr );
    pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_DETACHED );
    pthread_create( &b->io_thread_id, NULL, &io_thread_func, b );
    pthread_attr_destroy( &attr );
}

//////////////////////////////////////////////////
// Halt MPEG data stream to/from a file.

static void stop_stream( struct BOARD *b )
{
    // If stream i/o thread is running ...
    if ( b->mode != MODE_IDLE )
    {
	// Request thread termination.
	b->stop_rq = 1;

	// Wait for the thread to terminate
	while ( b->mode != MODE_IDLE )
	    sleep( 1 );
    }
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////  MPEG STREAM I/O THREAD  //////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////

char mpeg_buf[MPEG_BUF_SIZE];


/////////////////////////////////////////////////////
// Copy a block of MPEG data from board to file.
// Returns number of bytes copied.

static int mpeg_to_file( struct BOARD *b, FILE *file )
{
    int		len;
	struct v4l2_buffer buf;
    char *pbuf;
    int fd;

    fd = b->handle;
    memset(&buf, 0, sizeof(buf));
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    if (-1 == ioctl (fd, VIDIOC_DQBUF, &buf)) {
        switch (errno) {
        case EAGAIN:
            return 0;
        case EIO:
        default:
            printf("VIDIOC_DQBUF error");
            return errno;
        }
    }
    //assert (buf.index < n_buffers);
    pbuf = buffers[buf.index].start;
    len = buf.bytesused;

    if (len == -1) {
        switch (errno) {
        case EAGAIN:
            return 0;
        default:
            break;
        }
    }
    if ( len <= 0 ) {
        printf( "No signal\n" );
        prompt();
        b->stop_rq = 1;
        return 0;
    }
    //printf("mpeg buf %x %x %x %x\n", pbuf[0], pbuf[1], pbuf[2], pbuf[3]);
    // Write MPEG data to file.
    fwrite(pbuf, 1, len, file );
    
    if (-1 == ioctl (fd, VIDIOC_QBUF, &buf)) {
        printf("VIDIOC_QBUF error");
        prompt();
        len = 0;
    }
    // Return number of bytes transferred.
    return len;
}

/////////////////////////////////////////////////////
// Copy a block of MPEG data from file to board.
// Returns number of bytes copied.

static int mpeg_from_file( struct BOARD *b, FILE *file )
{
    //char	mpeg_buf[MPEG_BUF_SIZE];
    int		len;
    int		remlen;
    int		nwritten;
    int         j;

    // Attempt to read MPEG data from file.
    len = fread( mpeg_buf, 1, MPEG_BUF_SIZE, file );

    // Request Stream Halt and abort if end of file reached.
    if (len == 0) {
        // wait for decode to stop
        for (j = 0; j < 20; j++) {
            int rc;
            if (b->stop_rq)
                break;
            rc = ioctl( b->handle_decode, S2226_VIDIOC_DECODE_STATE, 0);
            if (rc == 0)
                break;
            sleep(1);
        }
        // let decoder finish
        if (!b->stop_rq)
            sleep(5);
        printf( "End of file reached.\n" );
        prompt();
        b->stop_rq = 1;
        return 0;
    }

    // Repeat until all MPEG bytes have been transferred to the board ...
    for ( remlen = len; remlen > 0; )
    {
	// Attempt to write MPEG data to board.
	nwritten = write( b->handle_decode, mpeg_buf, remlen );

	//  Request Stream Halt and abort if we timed out while writing MPEG data to board.
	if ( nwritten < 0 )
	{
	    printf( "Player error %d\n", b->handle );

	    prompt();
	    continue;
//	    b->stop_rq = 1;
//	    return 0;
	}

	// Compute remaining byte count.
	remlen -= nwritten;
    }

    // Return number of bytes transferred.
    return len;
}
#define    S2226_INPUT_COMPOSITE_0     0
#define    S2226_INPUT_SVIDEO_0        1
#define    S2226_INPUT_COMPOSITE_1     2
#define    S2226_INPUT_SVIDEO_1        3
#define    S2226_INPUT_SD_COLORBARS    4
#define    S2226_INPUT_720P_COLORBARS  5
#define    S2226_INPUT_1080I_COLORBARS 6
#define    S2226_INPUT_SDI_SD          7
#define    S2226_INPUT_SDI_720P_5994   8 
#define    S2226_INPUT_SDI_720P_60     9
#define    S2226_INPUT_SDI_1080I_50    10  
#define    S2226_INPUT_SDI_1080I_5994  11
#define    S2226_INPUT_SDI_1080I_60    12
#define    S2226_INPUT_SDI_720P_50     13
#define    S2226_INPUT_SDI_720P_24     14
#define    S2226_INPUT_SDI_720P_2398   15
#define    S2226_INPUT_SDI_1080P_24    16
#define    S2226_INPUT_SDI_1080P_2398  17


#define S2226_DECODE_480I       0
#define S2226_DECODE_576I       1
#define S2226_DECODE_1080I_60   2
#define S2226_DECODE_1080I_5994 3
#define S2226_DECODE_1080I_50   4
#define S2226_DECODE_720P_60    5
#define S2226_DECODE_720P_5994  6
#define S2226_DECODE_720P_50    7
#define S2226_DECODE_720P_24    8
#define S2226_DECODE_720P_2398  9
#define S2226_DECODE_1080P_24   10
#define S2226_DECODE_1080P_2398 11

// old == old input INPUT_COMP0_480I, INPUT_COMP1_480I, etc
// v4l == new V4L index 
// std == -1 (no standard call required), 0 --NTSC, 1 --PAL
void legacy_to_v4l_input(int old, int *v4l, int *std)
{
    switch (old) {
    case  INPUT_COMP0_480I:
        *std = 0;
        *v4l = S2226_INPUT_COMPOSITE_0;
        break;
    case INPUT_COMP0_576I:
        *std = 1;
        *v4l = S2226_INPUT_COMPOSITE_0;
        break;
    case INPUT_SVIDEO0_480I:
        *std = 0;
        *v4l = S2226_INPUT_SVIDEO_0;
        break;
    case INPUT_SVIDEO0_576I:
        *std = 0;
        *v4l = S2226_INPUT_SVIDEO_0;
        break;
    case INPUT_SDI_480I:
        *std = 0;
        *v4l = S2226_INPUT_SDI_SD;
        break;
    case INPUT_SDI_480I_CB:
        *std = 0;
        *v4l = S2226_INPUT_SD_COLORBARS;
        break;
    case INPUT_SDI_576I:
        *std = 1;
        *v4l = S2226_INPUT_SDI_SD;
        break;
    case INPUT_SDI_576I_CB:
        *std = 1;
        *v4l = S2226_INPUT_SD_COLORBARS;
        break;
    case INPUT_SDI_1080I_50:
        *std = -1; // don't care
        *v4l = S2226_INPUT_SDI_1080I_50;
        break;
    case INPUT_SDI_1080I_5994:
        *std = -1; // don't care
        *v4l = S2226_INPUT_SDI_1080I_5994;
        break;
    case INPUT_SDI_1080I_60:
        *std = -1; // don't care
        *v4l = S2226_INPUT_SDI_1080I_60;
        break;
    case INPUT_SDI_1080I_60_CB:
        *std = 0;
        *v4l = S2226_INPUT_1080I_COLORBARS;
        break;
    case INPUT_SDI_720P_50:
        *std = -1; // don't care
        *v4l = S2226_INPUT_SDI_720P_50;
        break;
    case INPUT_SDI_720P_5994:
        *std = -1; // don't care
        *v4l = S2226_INPUT_SDI_720P_5994;
        break;
    case INPUT_SDI_720P_60:
        *std = -1; // don't care
        *v4l = S2226_INPUT_SDI_720P_60;
        break;
    case INPUT_SDI_720P_60_CB:
        *std = 0;
        *v4l = S2226_INPUT_720P_COLORBARS;
        break;
#if 0
    case INPUT_H51_SD_480I:
    case INPUT_H51_HD_1080I_60:
    case INPUT_H51_HD_720P_60:
    case INPUT_H51_HD_720P_50:
    case INPUT_H51_SD_576I:
    case INPUT_H51_HD_1080I_50:
    case INPUT_H51_HD_1080I_5994:
    case INPUT_H51_HD_720P_5994:
#endif
    case INPUT_SDI_1080I_50_CB:
        *std = 1;
        *v4l = S2226_INPUT_1080I_COLORBARS;
        break;
    case INPUT_COMP1_480I:
        *std = 0;
        *v4l = S2226_INPUT_COMPOSITE_1;
        break;
    case INPUT_COMP1_576I:
        *std = 1;
        *v4l = S2226_INPUT_COMPOSITE_1;
        break;
    case INPUT_SVIDEO1_480I:
        *std = 0;
        *v4l = S2226_INPUT_SVIDEO_1;
        break;
    case INPUT_SVIDEO1_576I:
        *std = 1;
        *v4l = S2226_INPUT_SVIDEO_1;
        break;
#if 0
        // TODO
    case INPUT_SDI_1080I_5994_CB:
    case INPUT_SDI_720P_5994_CB:
#endif
    case INPUT_SDI_720P_50_CB:
        *std = 1;
        *v4l = S2226_INPUT_720P_COLORBARS;
        break;
#if 0
        // TODO
    case INPUT_SDI_1080P_30_CB:
    case INPUT_SDI_1080P_2997_CB:
    case INPUT_SDI_1080P_24_CB:
        *v4l = S2226_INPUT_SDI_1080P_24_CB;
        *std = -1;
        break;
    case INPUT_SDI_1080P_2398_CB:
    case INPUT_SDI_720P_24_CB:
    case INPUT_SDI_720P_2398_CB:
    case INPUT_SDI_1080P_30:
    case INPUT_SDI_1080P_2997:
#endif
    case INPUT_SDI_1080P_24:
        *std = -1;
        *v4l = S2226_INPUT_SDI_1080P_24;
        break;
    case INPUT_SDI_1080P_2398:
        *std = -1;
        *v4l = S2226_INPUT_SDI_1080P_2398;
        break;
    case INPUT_SDI_720P_24:
        *std = -1;
        *v4l = S2226_INPUT_SDI_720P_24;
        break;
    case INPUT_SDI_720P_2398:
        *std = -1;
        *v4l = S2226_INPUT_SDI_720P_24;
        break;
#if 0
    case INPUT_H51_HD_720P_24:
    case INPUT_H51_HD_720P_2398:
    case INPUT_H51_HD_1080P_24:
    case INPUT_H51_HD_1080P_2398:
#endif

    }

}

static int change_input(BOARD *b)
{
    int rc = 0;
    struct v4l2_input inp;
    int v4lidx;
    int is_pal;
    v4l2_std_id std;
    int lock;
    is_pal = -1;
    if ((b->mode == MODE_RECORD) || (b->mode == MODE_IDLE)) {
        //printf("selected %d, cur %d\n", b->settings.sel_video_input, b->settings.cur_video_input);
	    if (b->settings.sel_video_input != b->settings.cur_video_input || !b->settings.valid_input) {
            legacy_to_v4l_input(b->settings.sel_video_input, &v4lidx, &is_pal);
            inp.index = v4lidx;
            if (is_pal != -1) {
                std = is_pal ? V4L2_STD_PAL_B : V4L2_STD_NTSC_M;
                rc = ioctl(b->handle, VIDIOC_S_STD, &std);
                if (rc != 0) {
                    printf("could not set standard 0x%x\n", rc);
                }
            }
	    rc = ioctl(b->handle, VIDIOC_S_INPUT, &inp);
	    if (rc == 0)
		    b->settings.cur_video_input = b->settings.sel_video_input;
#ifndef NO_OVERLAYSNAP
            S2226_UpdateOverlay(b->settings.sel_video_input, b->handle);
#endif
	    rc = vid_status(b, &lock);
	    if ( rc != 0)
		    return rc;
	    if (lock != 1)
		    return -EINVAL;
	    }
    } if (b->mode == MODE_PLAY_HD_1080I_60) {
        inp.index = S2226_DECODE_1080I_60;
	    rc = ioctl( b->handle_decode, VIDIOC_S_INPUT, &inp);
	    if (rc == 0)
		    b->settings.cur_video_input = INPUT_H51_HD_1080I_60;
    } else if (b->mode == MODE_PLAY_HD_1080I_5994) {
        inp.index = S2226_DECODE_1080I_5994;
	    rc = ioctl( b->handle_decode, VIDIOC_S_INPUT, &inp);
	    if (rc == 0)
		    b->settings.cur_video_input = INPUT_H51_HD_1080I_5994;
    } else if (b->mode == MODE_PLAY_HD_1080I_50) {
        inp.index = S2226_DECODE_1080I_50;
	    rc = ioctl( b->handle_decode, VIDIOC_S_INPUT, &inp);
	    if (rc == 0)
		    b->settings.cur_video_input = INPUT_H51_HD_1080I_50;
    } else if (b->mode == MODE_PLAY_HD_720P_60) {
        inp.index = S2226_DECODE_1080I_50;
        rc = ioctl( b->handle_decode, VIDIOC_S_INPUT, &inp);
        if (rc == 0)
            b->settings.cur_video_input = INPUT_H51_HD_720P_60;
    } else if (b->mode == MODE_PLAY_HD_720P_5994) {
        inp.index = S2226_DECODE_720P_5994;
        rc = ioctl( b->handle_decode, VIDIOC_S_INPUT, &inp);
        if (rc == 0)
            b->settings.cur_video_input = INPUT_H51_HD_720P_5994;
    } else if (b->mode == MODE_PLAY_SD_480I) {
        inp.index = S2226_DECODE_480I;
        rc = ioctl( b->handle_decode, VIDIOC_S_INPUT, &inp);
        if (rc == 0)
            b->settings.cur_video_input = INPUT_H51_SD_480I;
    } else if (b->mode == MODE_PLAY_SD_576I) {
        inp.index = S2226_DECODE_576I;
        rc = ioctl( b->handle_decode, VIDIOC_S_INPUT, &inp);
        if (rc == 0)
            b->settings.cur_video_input = INPUT_H51_SD_576I;
    } else if (b->mode == MODE_PLAY_HD_720P_50) {
        inp.index = S2226_DECODE_720P_50;
        rc = ioctl( b->handle_decode, VIDIOC_S_INPUT, &inp);
        if (rc == 0)
            b->settings.cur_video_input = INPUT_H51_HD_720P_50;
    }
    return rc;
}


////////////////////////////////////////////////////////////////////
// I/O THREAD MAIN.  Read/write MPEG data until stream is stopped.
// The arg is a pointer to the device's BOARD object.

void *io_thread_func( void *arg )
{
    FILE		*file;			// MPEG data file.
    unsigned int	rlen;			// MPEG bytes transferred.
    time_t		starttime;		// Stream start time.
    BOARD		*b = (BOARD *)arg;	// Cache pointer to board object.
    start_param_t start;
    stop_param_t stop;

    // Select the function that will be used to move MPEG data between file and board.
    int (* mpeg_transfer)( struct BOARD *b, FILE *file ) = ( b->mode == MODE_RECORD ) ? mpeg_to_file : mpeg_from_file;

    // Initialize stream statistics.
    b->mpeg_count = 0;		// No MPEG bytes have been transferred.
    time( &starttime );		// Log the start time.
    if ((b->settings.valid_input == 0) && (b->mode == MODE_RECORD)) {
        b->mode = MODE_IDLE;
        printf( "input not set or not valid. please set input first\n");
        prompt();
        return NULL;
    }

    // Attempt to open the MPEG data file.  If successful ...
    if ( ( file = fopen( b->fname, ( b->mode == MODE_RECORD ) ? "wb" : "rb" ) ) != NULL )
    {
	// Report status.
	    int rc;
        start.idx = 0;
        rc = change_input(b);
        if (rc != 0) {
            if (rc == -EBUSY)
                printf( "start stream error: Stream busy(used by other process?) or no video signal\n");
            else
                printf( "start stream error(set input): Stream stopped\n");
            prompt();
            b->mode = MODE_IDLE;
            b->settings.cur_video_input = -1;
            b->settings.valid_input = 0;
            return NULL;
        }
        if (b->mode != MODE_RECORD) {
            rc = ioctl( b->handle_decode, S2226_VIDIOC_STARTDECODE, &start );
            if (rc != 0) {
                printf( "start stream error. Already running? %d\n", rc);
                prompt();
                b->mode = MODE_IDLE;
                b->settings.cur_video_input = -1;
                return NULL;
            }
            printf("Starting decode\n");
            prompt();
        } else {
            int type;
            int i;
            for (i = 0; i < n_buffers; ++i) {
                struct v4l2_buffer buf;
                memset(&buf, 0, sizeof(struct v4l2_buffer));
                buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                buf.memory = V4L2_MEMORY_MMAP;
                buf.index = i;
                if (-1 == ioctl(b->handle, VIDIOC_QBUF, &buf))
                    printf("qbuf error\n");
            }
            type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

            if (-1 == ioctl(b->handle, VIDIOC_STREAMON, &type)) {
                printf("failed to start stream!\n");
                prompt();
            } else {
                printf("start streamon\n");
                prompt();
            }
        }
        // Until stream termination is requested ...
        while (!b->stop_rq) {
            // Transfer a block of MPEG data between file and board.  If this is the first
            // transfer, the driver will automatically start the stream running.
            rlen = mpeg_transfer( b, file );
            // Collect statistical info.
            b->mpeg_count += rlen;
            // Allow other processes to execute if we are in non-blocking mode and no data was available.
            if ( ( rlen == 0 ) && !BLOCKING_MODE )	    {
                // Sleep for approximately 10 milliseconds.
                struct timespec tv;
                tv.tv_sec = 0;
                tv.tv_nsec = 10000000;
                nanosleep( &tv, &tv );
            }
        }
        // Halt the stream.
        stop.idx = 0;
        if (b->mode != MODE_RECORD) {
            printf("sending stop decode\n");
            rc = ioctl(b->handle_decode, S2226_VIDIOC_STOPDECODE, &stop);
            if (rc != 0) {
                printf( "stop stream error: %d\n", rc);
                prompt();
            }
        } else {
            int type;
            type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            ioctl(b->handle, VIDIOC_STREAMOFF, &type);
        }
        // Close the MPEG data file.
        fclose( file );
        
        // Report statistics.
        printf( "%u MPEG bytes transferred\n", b->mpeg_count );
        prompt();
    } else {
        // Indicate there was a problem opening the MPEG data file.
        printf( "file open error %s\n", b->fname);
        prompt();
    }

    // Indicate stream is stopped.
    b->mode = MODE_IDLE;

    printf( "Stream stopped\n");
    prompt();

    // Terminate thread.
    return 0;
}

