///////////////////////////////////////////////////////////////////////////
// File     : fpga2226.cpp
// Function : Encapsulation wrapper for FPGA register access functions
///////////////////////////////////////////////////////////////////////////

#include "s2226ioctl.h"
#include "fpga2226.h"
#include "platform.h"
#include "dh2226.h"
#include "image.h"
#include <unistd.h>
#include <time.h>
/*lint -e713 */ //Loss of precision (assignment) (unsigned int to int)
/*lint -e734 */ //Loss of precision (assignment) (31 bits to 7 bits)

#define GetRValue(rgb)      ((unsigned char)(rgb))
#define GetGValue(rgb)      ((unsigned char)((rgb) >> 8))
#define GetBValue(rgb)      ((unsigned char)((rgb) >>16))

#define MAX_SNAPWAIT_MSEC          500      // Max time to wait for SnapshotReadyForTransfer after a data request (in milliseconds).
#define SNAPWAIT_POLL_LOOP_MSEC     20      // Check this often while polling for SnapshotReadyForTransfer.

//////////////////////////////////////////////////////////////////////////
// Write to indexed register(s).
//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
// Write byte.
//////////////////////////////////////////////////////////////////////////
void WriteIndexedByte( MYHDEVICE handle, int addr, int value )
{
    // Note: These operations MUST be atomic.
    // Critical section was removed and 
    // a critical section taken around a large block of WriteIndexedByte
    // calls.  Any new individual calls to WriteIndexedByte MUST 
    // have a criticalsection to ensure atomic operation on 
    // multi-core CPUs
    s2226_write_data( handle, S2226_REG_WCINDEX, addr );
    s2226_write_data( handle, S2226_REG_WCDATA, value );
}

//////////////////////////////////////////////////////////////////////////
// Write word.
//////////////////////////////////////////////////////////////////////////
static void WriteIndexedWord( MYHDEVICE handle, int addr, int value )
{
    // Note: These operations MUST be atomic.
    // This critical section was removed and
    // a critical section taken around a large block of WriteIndexedByte
    // calls.  Any new individual calls to WriteIndexedByte MUST 
    // have a criticalsection to ensure atomic operation on 
    // multi-core CPUs
    WriteIndexedByte( handle, addr, value );
    WriteIndexedByte( handle, addr + 1, value >> 8 );
}

//////////////////////////////////////////////////////////////////////////
// Set overlay/black window position.
//////////////////////////////////////////////////////////////////////////
static void SetOBWindow( MYHDEVICE handle, int idx, int region, int top, int bottom, int left, int right )
{
    // Note: These operations MUST be atomic.
    // This critical section was removed
    // a critical section taken around a large block of WriteIndexedByte
    // calls.  Any new individual calls to WriteIndexedByte MUST 
    // have a criticalsection to ensure atomic operation on 
    // multi-core CPUs
    WriteIndexedWord( handle, idx, top | (region << 13) );
    WriteIndexedWord( handle, idx + 2, bottom - 1 );
    WriteIndexedWord( handle, idx + 4, left );
    WriteIndexedWord( handle, idx + 6, right - 1 );
}

//////////////////////////////////////////////////////////////////////////
//  API FUNCTIONS
//////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////
// Set overlay window position.
//////////////////////////////////////////////////////////////////////////
void SetOverlayWindowPosition( MYHDEVICE handle, int window_num, int region, int top, int bottom, int left, int right )
{
    SetOBWindow( handle, S2226_REG_LBLSB + (window_num << 3), region, top, bottom, left, right );
}

//////////////////////////////////////////////////////////////////////////
// Set black window position.
//////////////////////////////////////////////////////////////////////////
void SetBlackWindowPosition( MYHDEVICE handle, int window_num, int region, int top, int bottom, int left, int right )
{
    SetOBWindow( handle, S2226_REG_BWLBLSB + (window_num << 3), region, top, bottom, left, right );
}

static unsigned short RGB2YCrCb16_BT601loc( int R, int G, int B )
{
    int Y  = ( 16829*R + 33039*G +  6416*B + 1048576 ) >> ( 16 + 2 );
    int Cr = ( 28784*R - 24103*G -  4681*B + 8388608 ) >> ( 16 + 3 );
    int Cb = ( -9714*R - 19071*G + 28784*B + 8388608 ) >> ( 16 + 3 );
    unsigned short res;
    Y  = Y  > ( 235 >> 2 ) ? ( 235 >> 2 ) : Y;
    Cr = Cr > ( 240 >> 3 ) ? ( 240 >> 3 ) : Cr;
    Cb = Cb > ( 240 >> 3 ) ? ( 240 >> 3 ) : Cb;
    
    Y  = Y  < ( 16 >> 2 ) ? ( 16 >> 2 ) : Y;
    Cr = Cr < ( 16 >> 3 ) ? ( 16 >> 3 ) : Cr;
    Cb = Cb < ( 16 >> 3 ) ? ( 16 >> 3 ) : Cb;
    res = ( Y << 10 ) + ( Cr << 5 ) + Cb;
#if __BYTE_ORDER == __BIG_ENDIAN
    res = ((res >> 8) & 0xff) | ((res & 0xff) << 8);
#endif
    return res;
} // RGB2YCrCb16_BT601

//////////////////////////////////////////////////////////////////////////
// Set background window color.
//////////////////////////////////////////////////////////////////////////
void SetBlackWindowColor( MYHDEVICE handle, int color )    
{ 
    int R,G,B;
    int YCrCb;
    //inline static unsigned short RGB2YCrCb16( int R, int G, int B )
    R = GetRValue( color);
    B = GetBValue( color);
    G = GetGValue( color);

    YCrCb = RGB2YCrCb16_BT601loc( R, G, B); // Could be better if we check the HD/SD status and use BT.709 accordingly

    WriteIndexedWord( handle, S2226_REG_BWCOLOR0, YCrCb ); 
}

//////////////////////////////////////////////////////////////////////////
// Functions for writing to memory before writing to FPGA
//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
// Write byte.
//////////////////////////////////////////////////////////////////////////
void WriteAdrByteMem( char *pmem, int *pos, char addr, char value )
{
    pmem[*pos] = addr;
    *pos += 1;
    pmem[*pos] = value;
    *pos += 1;
}

//////////////////////////////////////////////////////////////////////////
// Write Indexed byte.
//////////////////////////////////////////////////////////////////////////
void WriteIndexedByteMem( char *pmem, int *pos, int addr, int value )
{
    pmem[*pos] = S2226_REG_WCINDEX;
    *pos += 1;
    pmem[*pos] = addr;
    *pos += 1;
    pmem[*pos] = S2226_REG_WCDATA;
    *pos += 1;
    pmem[*pos] = value;
    *pos += 1;
}

//////////////////////////////////////////////////////////////////////////
// Write word.
//////////////////////////////////////////////////////////////////////////
static void WriteIndexedWordMem( char *pmem, int *pos, int addr, int value )
{
    WriteIndexedByteMem( pmem, pos, addr, value );
    WriteIndexedByteMem( pmem, pos, addr + 1, value >> 8 );
}

//////////////////////////////////////////////////////////////////////////
// Set overlay/black window position.
//////////////////////////////////////////////////////////////////////////
static void SetOBWindowMem( char *pmem, int *pos, int idx, int region, int top, int bottom, int left, int right )
{
    WriteIndexedWordMem( pmem, pos, idx, top | (region << 13) );
    WriteIndexedWordMem( pmem, pos, idx + 2, bottom - 1 );
    WriteIndexedWordMem( pmem, pos, idx + 4, left );
    WriteIndexedWordMem( pmem, pos, idx + 6, right - 1 );
}

//////////////////////////////////////////////////////////////////////////
// Set overlay window position.
//////////////////////////////////////////////////////////////////////////
void SetOverlayWindowPositionMem( char *pmem, int *pos, int window_num, int region, int top, int bottom, int left, int right )
{
    SetOBWindowMem( pmem, pos, S2226_REG_LBLSB + (window_num << 3), region, top, bottom, left, right );
}

//////////////////////////////////////////////////////////////////////////
// Set black window position.
//////////////////////////////////////////////////////////////////////////
void SetBlackWindowPositionMem( char *pmem, int *pos, int window_num, int region, int top, int bottom, int left, int right )
{
    SetOBWindowMem( pmem, pos, S2226_REG_BWLBLSB + (window_num << 3), region, top, bottom, left, right );
}

//////////////////////////////////////////////////////////////////////////
// Set background window color.
//////////////////////////////////////////////////////////////////////////
void SetBlackWindowColorMem( char *pmem, int *pos , int color )    
{ 
    int R,G,B;
    int YCrCb;
    //inline static unsigned short RGB2YCrCb16( int R, int G, int B )
    R = GetRValue( color);
    B = GetBValue( color);
    G = GetGValue( color);
    YCrCb = RGB2YCrCb16_BT601( R, G, B); // Could be better if we check the HD/SD status and use BT.709 accordingly
    WriteIndexedWordMem( pmem, pos, S2226_REG_BWCOLOR0, YCrCb ); 
}

//////////////////////////////////////////////////////////////////////////
// Snapshot
//////////////////////////////////////////////////////////////////////////

/**
 * Acquire image.
 * Imports:
 * @param buf         - target image buffer.
 * @param bytecount   - maximum number of bytes to acquire.  Returns actual byte count transferred.
 * @param freezetime  - milliseconds that image will be frozen on display.
 * @param wait        - caller can wait if another thread is acquiring an image.
 * @return
 *   number of bytes transferred
 *   -1 if critical section locked
 *   -2 if syncronization loop times out
 */
int GetSnapshot(int handle, void *buf, int *bytecount, int freezetime, int wait )
{
    unsigned short ack;
    int rtnval = -2;    // return -2 if times out
    struct timespec starttime;
//  s2226_snapctrl(handle, 1);

    // Clear S2226_BIT_SNAP_DAV_N
    s2226_write_data( handle, S2226_REG_DEBUG, S2226_BIT_RESET_ERR ); // Reset bits 15-6 in S2226_REG_PLLFLAGS
    s2226_write_data( handle, S2226_REG_DEBUG, 0 );

    // Assert freeze request for specified time interval.
    s2226_write_data( handle, S2226_REG_FREEZE, 3 );  // freeze frame:bit-0, single frame freeze:bit-1

    // 4/20/2007: don't freeze the thread if not freezing.  eg. sleep(0) causes an unnecessary context switch
    if( freezetime) { 
        usleep( freezetime * 1000 );
    } else { 
        // freeze for a minimum value greater than the video frame rate is 
        // required by hardware
        usleep( 30 * 1000 );
    }

    // Make SURE the snapshot has been acquired before releasing freeze!
    //
    // S2226_BIT_SNAP_DAV_N gets set when the internal snapshot memory's
    // inverted Read_FIFO_Not_Empty transitions from 0 to 1.
    // 1) During reset. (Before/after snapshot capture and before/after
    // the freeze image replay, the FIFO is empty, thus SNAP_DAV_N=1
    // 2) During image capture, the FIFO reads data and SNAP_DAV_N=0
    // 3) After the capture the FIFO is reset and       SNAP_DAV_N=1
    // 4) S2226_BIT_SNAP_DAV_N now set. (INV_SNAP_DAV_N went 0 to 1)
    current_time(&starttime);
    do {
      s2226_read_addr_data( handle, S2226_REG_PLLFLAGS, &ack );
    } while ( !(ack & S2226_BIT_SNAP_DAV_N) && ( elapsed_time(&starttime) < 120));
    if (!(ack & S2226_BIT_SNAP_DAV_N)) {
        // Release freeze request and assert data request.
        s2226_write_data( handle, S2226_REG_FREEZE, 0 );
        return rtnval;
    }

    // Release freeze request and assert data request.
    s2226_write_data( handle, S2226_REG_FREEZE, 0 );

    // reset image sequence register
    s2226_read_addr_data( handle, S2226_REG_SNAPSTAT, &ack ); // Reset image data sequencer, Returns SNAPDAT read type
    s2226_write_data( handle, S2226_REG_SNAPREQ, 1 );
    // Repeat until image is transferred or time-out ...
    current_time(&starttime);

    do
    {
        // If image data is ready for transfer ...
        s2226_read_addr_data( handle, S2226_REG_SNAPACK, &ack ); // b0 = not rdempty in output fifo
        if ( ack )
        {
            // Burst transfer image data to application buffer.
            rtnval = s2226_BurstReadTransfer( handle, S2226_REG_SNAPDAT, buf, *bytecount ); 
            // Return number of bytes transferred.
            // Exit poll loop.
            break;
        }
        // Wait a short time.
        usleep( SNAPWAIT_POLL_LOOP_MSEC * 1000);
    } 
    while ( elapsed_time(&starttime) < MAX_SNAPWAIT_MSEC );
    // Negate data request.
    s2226_write_data( handle, S2226_REG_SNAPREQ, 0 );

    //// Enable next snapshot request.
//  s2226_snapctrl(handle, 0);
    // Indicate pass/fail.
    return rtnval;
} // GetSnapshot


//////////////////////////////////////////////////////////
// Write to overlay ram.
// Imports:
//  buf         - source buffer for overlay data.
//  bytecount   - number of bytes to write.
//  wait        - caller can wait if another thread is writing to overlay ram.
// Returns true if successful, false if lock failed.

int SetOverlayMem( MYHDEVICE handle, void *buf, int *bytecount, int wait, int arm_ver )
{
    // Reset the overlay address.
    s2226_write_data( handle, S2226_REG_OVACLR, 0 );
    // Burst the overlay data out to overlay ram.
    if (arm_ver < 0x52)
        s2226_BurstWriteTransfer(handle, S2226_REG_OVDATA, (unsigned char *) buf, *bytecount, 0);
  else {
      s2226_BurstWriteTransferFast(handle, S2226_REG_OVDATA, (unsigned char *) buf, *bytecount, 0);
  }

    // Reset the overlay address and flush DDR write fifo.
    s2226_write_data( handle, S2226_REG_OVACLR, 1 );

    // Enable next overlay write.
    // Indicate success.
    return TRUE;
}
